# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from .webkit_base import WebkitBase

WEBKIT_SAFARIVIEWSERVICE_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/SystemData/com.apple.SafariViewService/Library/WebKit/WebsiteData/",
]

class WebkitSafariViewService(WebkitBase):
    """This module looks extracts records from WebKit LocalStorage folders,
    and checks them against any provided list of suspicious domains."""

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def run(self):
        self._database_from_path(WEBKIT_SAFARIVIEWSERVICE_ROOT_PATHS)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import csv
import glob
import io
import logging
import os
import re

import simplejson as json

from .indicators import Indicators


class DatabaseNotFoundError(Exception):
    pass

class DatabaseCorruptedError(Exception):
    pass

class InsufficientPrivileges(Exception):
    pass

class MVTModule(object):
    """This class provides a base for all extraction modules."""

    enabled = True
    slug = None

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        """Initialize module.
        :param file_path: Path to the module's database file, if there is any.
        :param base_folder: Path to the base folder (backup or filesystem dump)
        :param output_folder: Folder where results will be stored
        :param fast_mode: Flag to enable or disable slow modules
        :param log: Handle to logger
        :param results: Provided list of results entries
        """
        self.file_path = file_path
        self.base_folder = base_folder
        self.output_folder = output_folder
        self.fast_mode = fast_mode
        self.log = log
        self.indicators = None
        self.results = results
        self.detected = []
        self.timeline = []
        self.timeline_detected = []

    @classmethod
    def from_json(cls, json_path, log=None):
        with open(json_path, "r") as handle:
            results = json.load(handle)
            if log:
                log.info("Loaded %d results from \"%s\"",
                         len(results), json_path)
            return cls(results=results, log=log)

    def get_slug(self):
        if self.slug:
            return self.slug

        sub = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", self.__class__.__name__)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", sub).lower()

    def _find_paths(self, root_paths):
        for root_path in root_paths:
            for found_path in glob.glob(os.path.join(self.base_folder, root_path)):
                if not os.path.exists(found_path):
                    continue
                yield found_path

    def load_indicators(self, file_path):
        self.indicators = Indicators(file_path, self.log)

    def check_indicators(self):
        """Check the results of this module against a provided list of
        indicators."""
        raise NotImplementedError

    def save_to_json(self):
        """Save the collected results to a json file.
        """
        if not self.output_folder:
            return

        name = self.get_slug()

        if self.results:
            results_file_name = f"{name}.json"
            results_json_path = os.path.join(self.output_folder, results_file_name)
            with open(results_json_path, "w") as handle:
                try:
                    json.dump(self.results, handle, indent=4)
                except Exception as e:
                    self.log.error("Unable to store results of module %s to file %s: %s",
                                   self.__class__.__name__, results_file_name, e)

        if self.detected:
            detected_file_name = f"{name}_detected.json"
            detected_json_path = os.path.join(self.output_folder, detected_file_name)
            with open(detected_json_path, "w") as handle:
                json.dump(self.detected, handle, indent=4)

    def serialize(self, record):
        raise NotImplementedError

    def to_timeline(self):
        """Convert results into a timeline.
        """
        for result in self.results:
            record = self.serialize(result)
            if record:
                if type(record) == list:
                    self.timeline.extend(record)
                else:
                    self.timeline.append(record)

        for detected in self.detected:
            record = self.serialize(detected)
            if record:
                if type(record) == list:
                    self.timeline_detected.extend(record)
                else:
                    self.timeline_detected.append(record)

        # De-duplicate timeline entries.
        self.timeline = self.timeline_deduplicate(self.timeline)
        self.timeline_detected = self.timeline_deduplicate(self.timeline_detected)

    def timeline_deduplicate(self, timeline):
        """Serialize entry as JSON to deduplicate repeated entries"""
        timeline_set = set()
        for record in timeline:
            timeline_set.add(json.dumps(record, sort_keys=True))
        return [json.loads(record) for record in timeline_set]

    def run(self):
        """Run the main module procedure.
        """
        raise NotImplementedError


def run_module(module):
    module.log.info("Running module %s...", module.__class__.__name__)

    try:
        module.run()
    except NotImplementedError:
        module.log.exception("The run() procedure of module %s was not implemented yet!",
                             module.__class__.__name__)
    except InsufficientPrivileges as e:
        module.log.info("Insufficient privileges for module %s: %s", module.__class__.__name__, e)
    except DatabaseNotFoundError as e:
        module.log.info("There might be no data to extract by module %s: %s",
                        module.__class__.__name__, e)
    except DatabaseCorruptedError as e:
        module.log.error("The %s module database seems to be corrupted and recovery failed: %s",
                         module.__class__.__name__, e)
    except Exception as e:
        module.log.exception("Error in running extraction from module %s: %s",
                             module.__class__.__name__, e)
    else:
        try:
            module.check_indicators()
        except NotImplementedError:
            module.log.info("The %s module does not support checking for indicators",
                            module.__class__.__name__)
            pass
        else:
            if module.indicators and not module.detected:
                module.log.info("The %s module produced no detections!",
                                module.__class__.__name__)

        try:
            module.to_timeline()
        except NotImplementedError:
            pass

        module.save_to_json()


def save_timeline(timeline, timeline_path):
    """Save the timeline in a csv file.
    :param timeline: List of records to order and store.
    :param timeline_path: Path to the csv file to store the timeline to.
    """
    with io.open(timeline_path, "a+", encoding="utf-8") as handle:
        csvoutput = csv.writer(handle, delimiter=",", quotechar="\"")
        csvoutput.writerow(["UTC Timestamp", "Plugin", "Event", "Description"])
        for event in sorted(timeline, key=lambda x: x["timestamp"] if x["timestamp"] is not None else ""):
            csvoutput.writerow([
                event["timestamp"],
                event["module"],
                event["event"],
                event["data"],
            ])



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from .cache_files import CacheFiles
from .calls import Calls
from .chrome_favicon import ChromeFavicon
from .chrome_history import ChromeHistory
from .contacts import Contacts
from .filesystem import Filesystem
from .firefox_favicon import FirefoxFavicon
from .firefox_history import FirefoxHistory
from .idstatuscache import IDStatusCache
from .interactionc import InteractionC
from .locationd import LocationdClients
from .manifest import Manifest
from .net_datausage import Datausage
from .net_netusage import Netusage
from .safari_browserstate import SafariBrowserState
from .safari_favicon import SafariFavicon
from .safari_history import SafariHistory
from .sms import SMS
from .sms_attachments import SMSAttachments
from .version_history import IOSVersionHistory
from .webkit_indexeddb import WebkitIndexedDB
from .webkit_localstorage import WebkitLocalStorage
from .webkit_resource_load_statistics import WebkitResourceLoadStatistics
from .webkit_safariviewservice import WebkitSafariViewService
from .webkit_session_resource_log import WebkitSessionResourceLog
from .whatsapp import Whatsapp

BACKUP_MODULES = [SafariBrowserState, SafariHistory, Datausage, SMS, SMSAttachments,
                  ChromeHistory, ChromeFavicon, WebkitSessionResourceLog,
                  WebkitResourceLoadStatistics, Calls, IDStatusCache, LocationdClients,
                  InteractionC, FirefoxHistory, FirefoxFavicon, Contacts, Manifest, Whatsapp]

FS_MODULES = [IOSVersionHistory, SafariHistory, SafariFavicon, SafariBrowserState,
              WebkitIndexedDB, WebkitLocalStorage, WebkitSafariViewService,
              WebkitResourceLoadStatistics, WebkitSessionResourceLog,
              Datausage, Netusage, ChromeHistory,
              ChromeFavicon, Calls, IDStatusCache, SMS, SMSAttachments,
              LocationdClients, InteractionC, FirefoxHistory, FirefoxFavicon,
              Contacts, CacheFiles, Whatsapp, Filesystem]



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import sqlite3

from mvt.common.utils import (convert_chrometime_to_unix,
                              convert_timestamp_to_iso)

from .base import IOSExtraction

CHROME_HISTORY_BACKUP_IDS = [
    "faf971ce92c3ac508c018dce1bef2a8b8e9838f1",
]

# TODO: Confirm Chrome database path.
CHROME_HISTORY_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/Library/Application Support/Google/Chrome/Default/History",
]

class ChromeHistory(IOSExtraction):
    """This module extracts all Chome visits."""

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def serialize(self, record):
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "visit",
            "data": f"{record['id']} - {record['url']} (visit ID: {record['visit_id']}, redirect source: {record['redirect_source']})"
        }

    def check_indicators(self):
        if not self.indicators:
            return

        for result in self.results:
            if self.indicators.check_domain(result["url"]):
                self.detected.append(result)

    def run(self):
        self._find_ios_database(backup_ids=CHROME_HISTORY_BACKUP_IDS, root_paths=CHROME_HISTORY_ROOT_PATHS)
        self.log.info("Found Chrome history database at path: %s", self.file_path)

        conn = sqlite3.connect(self.file_path)
        cur = conn.cursor()
        cur.execute("""
            SELECT
                urls.id,
                urls.url,
                visits.id,
                visits.visit_time,
                visits.from_visit
            FROM urls
            JOIN visits ON visits.url = urls.id
            ORDER BY visits.visit_time;
        """)

        for item in cur:
            self.results.append(dict(
                id=item[0],
                url=item[1],
                visit_id=item[2],
                timestamp=item[3],
                isodate=convert_timestamp_to_iso(convert_chrometime_to_unix(item[3])),
                redirect_source=item[4],
            ))

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d history items", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import io
import sqlite3

import biplist

from mvt.common.utils import (convert_mactime_to_unix,
                              convert_timestamp_to_iso, keys_bytes_to_string)

from .base import IOSExtraction

SAFARI_BROWSER_STATE_BACKUP_IDS = [
    "3a47b0981ed7c10f3e2800aa66bac96a3b5db28e",
]
SAFARI_BROWSER_STATE_ROOT_PATHS = [
    "private/var/mobile/Library/Safari/BrowserState.db",
    "private/var/mobile/Containers/Data/Application/*/Library/Safari/BrowserState.db",
]

class SafariBrowserState(IOSExtraction):
    """This module extracts all Safari browser state records."""

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def serialize(self, record):
        return {
            "timestamp": record["last_viewed_timestamp"],
            "module": self.__class__.__name__,
            "event": "tab",
            "data": f"{record['tab_title']} - {record['tab_url']}"
        }

    def check_indicators(self):
        if not self.indicators:
            return

        for result in self.results:
            if "tab_url" in result and self.indicators.check_domain(result["tab_url"]):
                self.detected.append(result)
                continue

            if not "session_data" in result:
                continue

            for session_entry in result["session_data"]:
                if "entry_url" in session_entry and self.indicators.check_domain(session_entry["entry_url"]):
                    self.detected.append(result)

    def run(self):
        self._find_ios_database(backup_ids=SAFARI_BROWSER_STATE_BACKUP_IDS,
                                root_paths=SAFARI_BROWSER_STATE_ROOT_PATHS)
        self.log.info("Found Safari browser state database at path: %s", self.file_path)

        conn = sqlite3.connect(self.file_path)

        # Fetch valid icon cache.
        cur = conn.cursor()
        cur.execute("""SELECT
                tabs.title,
                tabs.url,
                tabs.user_visible_url,
                tabs.last_viewed_time,
                tab_sessions.session_data
            FROM tabs
            JOIN tab_sessions ON tabs.uuid = tab_sessions.tab_uuid
            ORDER BY tabs.last_viewed_time;""")

        session_history_count = 0
        for item in cur:
            session_entries = []

            if item[4]:
                # Skip a 4 byte header before the plist content.
                session_plist = item[4][4:]
                session_data = biplist.readPlist(io.BytesIO(session_plist))
                session_data = keys_bytes_to_string(session_data)

                if "SessionHistoryEntries" in session_data["SessionHistory"]:
                    for session_entry in session_data["SessionHistory"]["SessionHistoryEntries"]:
                        session_history_count += 1
                        session_entries.append(dict(
                            entry_title=session_entry["SessionHistoryEntryOriginalURL"],
                            entry_url=session_entry["SessionHistoryEntryURL"],
                            data_length=len(session_entry["SessionHistoryEntryData"]) if "SessionHistoryEntryData" in session_entry else 0,
                        ))

            self.results.append(dict(
                tab_title=item[0],
                tab_url=item[1],
                tab_visible_url=item[2],
                last_viewed_timestamp=convert_timestamp_to_iso(convert_mactime_to_unix(item[3])),
                session_data=session_entries,
            ))

        self.log.info("Extracted a total of %d tab records and %d session history entries",
                      len(self.results), session_history_count)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import json
import logging
import os

import pkg_resources
from tqdm import tqdm

from mvt.common.utils import get_sha256_from_file_path

from .modules.adb.base import AndroidExtraction

log = logging.getLogger(__name__)

# TODO: Would be better to replace tqdm with rich.progress to reduce
#       the number of dependencies. Need to investigate whether
#       it's possible to have a similar callback system.
class PullProgress(tqdm):
    """PullProgress is a tqdm update system for APK downloads."""

    def update_to(self, file_name, current, total):
        if total is not None:
            self.total = total
        self.update(current - self.n)


class Package:
    """Package indicates a package name and all the files associated with it."""

    def __init__(self, name, files=None):
        self.name = name
        self.files = files or []


class DownloadAPKs(AndroidExtraction):
    """DownloadAPKs is the main class operating the download of APKs
    from the device."""

    def __init__(self, output_folder=None, all_apks=False, packages=None):
        """Initialize module.
        :param output_folder: Path to the folder where data should be stored
        :param all_apks: Boolean indicating whether to download all packages
                         or filter known-goods
        :param packages: Provided list of packages, typically for JSON checks
        """
        super().__init__(file_path=None, base_folder=None,
                         output_folder=output_folder)

        self.output_folder_apk = None
        self.packages = packages or []
        self.all_apks = all_apks

        self._safe_packages = []

    @classmethod
    def from_json(cls, json_path):
        """Initialize this class from an existing packages.json file.
        :param json_path: Path to the packages.json file to parse.
        """
        with open(json_path, "r") as handle:
            data = json.load(handle)

            packages = []
            for entry in data:
                package = Package(entry["name"], entry["files"])
                packages.append(package)

            return cls(packages=packages)

    def _load_safe_packages(self):
        """Load known-good package names.
        """
        safe_packages_path = os.path.join("data", "safe_packages.txt")
        safe_packages_string = pkg_resources.resource_string(__name__, safe_packages_path)
        safe_packages_list = safe_packages_string.decode("utf-8").split("\n")
        self._safe_packages.extend(safe_packages_list)

    def _clean_output(self, output):
        """Clean adb shell command output.
        :param output: Command output to clean.
        """
        return output.strip().replace("package:", "")

    def get_packages(self):
        """Retrieve package names from the device using adb.
        """
        log.info("Retrieving package names ...")

        if not self.all_apks:
            self._load_safe_packages()

        output = self._adb_command("pm list packages")
        total = 0
        for line in output.split("\n"):
            package_name = self._clean_output(line)
            if package_name == "":
                continue

            total += 1

            if not self.all_apks and package_name in self._safe_packages:
                continue

            if package_name not in self.packages:
                self.packages.append(Package(package_name))

        log.info("There are %d packages installed on the device. I selected %d for inspection.",
                 total, len(self.packages))

    def pull_package_file(self, package_name, remote_path):
        """Pull files related to specific package from the device.
        :param package_name: Name of the package to download
        :param remote_path: Path to the file to download
        :returns: Path to the local copy
        """
        log.info("Downloading %s ...", remote_path)

        file_name = ""
        if "==/" in remote_path:
            file_name = "_" + remote_path.split("==/")[1].replace(".apk", "")

        local_path = os.path.join(self.output_folder_apk,
                                  f"{package_name}{file_name}.apk")
        name_counter = 0
        while True:
            if not os.path.exists(local_path):
                break

            name_counter += 1
            local_path = os.path.join(self.output_folder_apk,
                                      f"{package_name}{file_name}_{name_counter}.apk")

        try:
            with PullProgress(unit='B', unit_divisor=1024, unit_scale=True,
                              miniters=1) as pp:
                self._adb_download(remote_path, local_path,
                                   progress_callback=pp.update_to)
        except Exception as e:
            log.exception("Failed to pull package file from %s: %s",
                          remote_path, e)
            self._adb_reconnect()
            return None

        return local_path

    def pull_packages(self):
        """Download all files of all selected packages from the device.
        """
        log.info("Starting extraction of installed APKs at folder %s", self.output_folder)

        if not os.path.exists(self.output_folder):
            os.mkdir(self.output_folder)

        log.info("Downloading packages from device. This might take some time ...")

        self.output_folder_apk = os.path.join(self.output_folder, "apks")
        if not os.path.exists(self.output_folder_apk):
            os.mkdir(self.output_folder_apk)

        total_packages = len(self.packages)
        counter = 0
        for package in self.packages:
            counter += 1

            log.info("[%d/%d] Package: %s", counter, total_packages, package.name)

            try:
                output = self._adb_command(f"pm path {package.name}")
                output = self._clean_output(output)
                if not output:
                    continue
            except Exception as e:
                log.exception("Failed to get path of package %s: %s", package.name, e)
                self._adb_reconnect()
                continue

            # Sometimes the package path contains multiple lines for multiple apks.
            # We loop through each line and download each file.
            for path in output.split("\n"):
                device_path = path.strip()
                file_path = self.pull_package_file(package.name, device_path)
                if not file_path:
                    continue

                # We add the apk metadata to the package object.
                package.files.append({
                    "path": device_path,
                    "local_name": file_path,
                    "sha256": get_sha256_from_file_path(file_path),
                })

    def save_json(self):
        """Save the results to the package.json file.
        """
        json_path = os.path.join(self.output_folder, "packages.json")
        packages = []
        for package in self.packages:
            packages.append(package.__dict__)

        with open(json_path, "w") as handle:
            json.dump(packages, handle, indent=4)

    def run(self):
        """Run all steps of fetch-apk.
        """
        self._adb_connect()
        self.get_packages()
        self.pull_packages()
        self.save_json()
        self._adb_disconnect()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import sqlite3

from mvt.common.url import URL
from mvt.common.utils import convert_mactime_to_unix, convert_timestamp_to_iso

from .base import IOSExtraction

SAFARI_HISTORY_BACKUP_IDS = [
    "e74113c185fd8297e140cfcf9c99436c5cc06b57",
    "1a0e7afc19d307da602ccdcece51af33afe92c53",
]
SAFARI_HISTORY_ROOT_PATHS = [
    "private/var/mobile/Library/Safari/History.db",
    "private/var/mobile/Containers/Data/Application/*/Library/Safari/History.db",
]

class SafariHistory(IOSExtraction):
    """This module extracts all Safari visits and tries to detect potential
    network injection attacks."""

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def serialize(self, record):
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "safari_history",
            "data": f"Safari visit to {record['url']} (ID: {record['id']}, Visit ID: {record['visit_id']})",
        }

    def _find_injections(self):
        for result in self.results:
            # We presume injections only happen on HTTP visits.
            if not result["url"].lower().startswith("http://"):
                continue

            # If there is no destination, no redirect happened.
            if not result["redirect_destination"]:
                continue

            origin_domain = URL(result["url"]).domain

            # We loop again through visits in order to find redirect record.
            for redirect in self.results:
                if redirect["visit_id"] != result["redirect_destination"]:
                    continue

                redirect_domain = URL(redirect["url"]).domain
                # If the redirect destination is the same domain as the origin,
                # it's most likely an HTTPS upgrade.
                if origin_domain == redirect_domain:
                    continue

                self.log.info("Found HTTP redirect to different domain: \"%s\" -> \"%s\"",
                         origin_domain, redirect_domain)

                redirect_time = convert_mactime_to_unix(redirect["timestamp"])
                origin_time = convert_mactime_to_unix(result["timestamp"])
                elapsed_time = redirect_time - origin_time
                elapsed_ms = elapsed_time.microseconds / 1000

                if elapsed_time.seconds == 0:
                    self.log.warning("Redirect took less than a second! (%d milliseconds)", elapsed_ms)

    def check_indicators(self):
        self._find_injections()

        if not self.indicators:
            return

        for result in self.results:
            if self.indicators.check_domain(result["url"]):
                self.detected.append(result)

    def run(self):
        self._find_ios_database(backup_ids=SAFARI_HISTORY_BACKUP_IDS, root_paths=SAFARI_HISTORY_ROOT_PATHS)
        self.log.info("Found Safari history database at path: %s", self.file_path)

        conn = sqlite3.connect(self.file_path)
        cur = conn.cursor()
        cur.execute("""
            SELECT
                history_items.id,
                history_items.url,
                history_visits.id,
                history_visits.visit_time,
                history_visits.redirect_source,
                history_visits.redirect_destination
            FROM history_items
            JOIN history_visits ON history_visits.history_item = history_items.id
            ORDER BY history_visits.visit_time;
        """)

        items = []
        for item in cur:
            items.append(dict(
                id=item[0],
                url=item[1],
                visit_id=item[2],
                timestamp=item[3],
                isodate=convert_timestamp_to_iso(convert_mactime_to_unix(item[3])),
                redirect_source=item[4],
                redirect_destination=item[5]
            ))

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d history items", len(items))
        self.results = items



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import sqlite3

from mvt.common.utils import convert_mactime_to_unix, convert_timestamp_to_iso

from .base import IOSExtraction

CALLS_BACKUP_IDS = [
    "5a4935c78a5255723f707230a451d79c540d2741",
]
CALLS_ROOT_PATHS = [
    "private/var/mobile/Library/CallHistoryDB/CallHistory.storedata"
]

class Calls(IOSExtraction):
    """This module extracts phone calls details"""

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def serialize(self, record):
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "call",
            "data": f"From {record['number']} using {record['provider']} during {record['duration']} seconds"
        }

    def run(self):
        self._find_ios_database(backup_ids=CALLS_BACKUP_IDS, root_paths=CALLS_ROOT_PATHS)
        self.log.info("Found Calls database at path: %s", self.file_path)

        conn = sqlite3.connect(self.file_path)
        cur = conn.cursor()
        cur.execute("""
            SELECT
                ZDATE, ZDURATION, ZLOCATION, ZADDRESS, ZSERVICE_PROVIDER
                FROM ZCALLRECORD;
        """)
        names = [description[0] for description in cur.description]

        for entry in cur:
            self.results.append({
                "isodate": convert_timestamp_to_iso(convert_mactime_to_unix(entry[0])),
                "duration": entry[1],
                "location": entry[2],
                "number": entry[3].decode("utf-8") if entry[3] and entry[3] is bytes else entry[3],
                "provider": entry[4]
            })

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d calls", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import datetime
import os
import sqlite3

from mvt.common.utils import convert_mactime_to_unix, convert_timestamp_to_iso

from .base import IOSExtraction

WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH = "Library/WebKit/WebsiteData/ResourceLoadStatistics/observations.db"
WEBKIT_RESOURCELOADSTATICS_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/Library/WebKit/WebsiteData/ResourceLoadStatistics/observations.db",
    "private/var/mobile/Containers/Data/Application/*/SystemData/com.apple.SafariViewService/Library/WebKit/WebsiteData/observations.db",
]

class WebkitResourceLoadStatistics(IOSExtraction):
    """This module extracts records from WebKit ResourceLoadStatistics observations.db.
    """
    # TODO: Add serialize().

    def __init__(self, file_path=None, base_folder=None, output_folder=None,
                 fast_mode=False, log=None, results=[]):
        super().__init__(file_path=file_path, base_folder=base_folder,
                         output_folder=output_folder, fast_mode=fast_mode,
                         log=log, results=results)

    def check_indicators(self):
        if not self.indicators:
            return

        self.detected = {}
        for key, items in self.results.items():
            for item in items:
                if self.indicators.check_domain(item["registrable_domain"]):
                    if key not in self.detected:
                        self.detected[key] = [item,]
                    else:
                        self.detected[key].append(item)

    def _process_observations_db(self, db_path, key):
        self.log.info("Found WebKit ResourceLoadStatistics observations.db file at path %s", db_path)

        if self._is_database_malformed(db_path):
            self._recover_database(db_path)

        conn = sqlite3.connect(db_path)
        cur = conn.cursor()

        try:
            cur.execute("SELECT * from ObservedDomains;")
        except sqlite3.OperationalError:
            return

        if not key in self.results:
            self.results[key] = []

        for row in cur:
            self.results[key].append(dict(
                domain_id=row[0],
                registrable_domain=row[1],
                last_seen=row[2],
                had_user_interaction=bool(row[3]),
                # TODO: Fix isodate.
                last_seen_isodate=convert_timestamp_to_iso(datetime.datetime.utcfromtimestamp(int(row[2]))),
            ))

        if len(self.results[key]) > 0:
            self.log.info("Extracted a total of %d records from %s", len(self.results[key]), db_path)

    def run(self):
        self.results = {}

        if self.is_backup:
            manifest_db_path = os.path.join(self.base_folder, "Manifest.db")
            if not os.path.exists(manifest_db_path):
                self.log.info("Unable to search for WebKit observations.db files in backup because of missing Manifest.db")
                return

            try:
                conn = sqlite3.connect(manifest_db_path)
                cur = conn.cursor()
                cur.execute("SELECT fileID, domain FROM Files WHERE relativePath = ?;", (WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH,))
            except Exception as e:
                self.log.error("Unable to search for WebKit observations.db files in backup because of failed query to Manifest.db: %s", e)

            for row in cur:
                file_id = row[0]
                domain = row[1]
                db_path = os.path.join(self.base_folder, file_id[0:2], file_id)
                if os.path.exists(db_path):
                    self._process_observations_db(db_path=db_path, key=f"{domain}/{WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH}")
        elif self.is_fs_dump:
            for db_path in self._find_paths(WEBKIT_RESOURCELOADSTATICS_ROOT_PATHS):
                self._process_observations_db(db_path=db_path, key=os.path.relpath(db_path, self.base_folder))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021 The MVT Project Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging

import requests
from rich.console import Console
from rich.progress import track
from rich.table import Table
from rich.text import Text

log = logging.getLogger(__name__)

def get_virustotal_report(hashes):
    apikey = "233f22e200ca5822bd91103043ccac138b910db79f29af5616a9afe8b6f215ad"
    url = f"https://www.virustotal.com/partners/sysinternals/file-reports?apikey={apikey}"

    items = []
    for sha256 in hashes:
        items.append({
            "autostart_location": "",
            "autostart_entry": "",
            "hash": sha256,
            "local_name": "",
            "creation_datetime": "",
        })
    headers = {"User-Agent": "VirusTotal", "Content-Type": "application/json"}
    res = requests.post(url, headers=headers, json=items)

    if res.status_code == 200:
        report = res.json()
        return report["data"]
    else:
        log.error("Unexpected response from VirusTotal: %s", res.status_code)
        return None

def virustotal_lookup(packages):
    log.info("Looking up all extracted files on VirusTotal (www.virustotal.com)")

    unique_hashes = []
    for package in packages:
        for file in package.files:
            if file["sha256"] not in unique_hashes:
                unique_hashes.append(file["sha256"])

    total_unique_hashes = len(unique_hashes)

    detections = {}
    def virustotal_query(batch):
        report = get_virustotal_report(batch)
        if not report:
            return

        for entry in report:
            if entry["hash"] not in detections and entry["found"] is True:
                detections[entry["hash"]] = entry["detection_ratio"]

    batch = []
    for i in track(range(total_unique_hashes), description=f"Looking up {total_unique_hashes} files..."):
        file_hash = unique_hashes[i]
        batch.append(file_hash)
        if len(batch) == 25:
            virustotal_query(batch)
            batch = []

    if batch:
        virustotal_query(batch)

    table = Table(title="VirusTotal Packages Detections")
    table.add_column("Package name")
    table.add_column("File path")
    table.add_column("Detections")

    for package in packages:
        for file in package.files:
            row = [package.name, file["local_name"]]

            if file["sha256"] in detections:
                detection = detections[file["sha256"]]
                positives = detection.split("/")[0]
                if int(positives) > 0:
                    row.append(Text(detection, "red bold"))
                else:
                    row.append(detection)
            else:
                row.append("not found")

            table.add_row(*row)

    console = Console()
    console.print(table)
